@@ -1,6 +1,6 @@
 /*
  * Copyright (c) [2011-2015] Novell, Inc.
- * Copyright (c) [2016-2018] SUSE LLC
+ * Copyright (c) [2016-2019] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -443,6 +443,28 @@ command_list(ProxySnappers* snappers, ProxySnapper*)
 void
 list_from_one_config(ProxySnapper* snapper, ListMode list_mode, bool show_used_space)
 {
+    const ProxySnapshots& snapshots = snapper->getSnapshots();
+
+    ProxySnapshots::const_iterator default_snapshot = snapshots.end();
+    ProxySnapshots::const_iterator active_snapshot = snapshots.end();
+
+    try
+    {
+	default_snapshot = snapshots.getDefault();
+	active_snapshot = snapshots.getActive();
+    }
+    catch (const DBus::ErrorException& e)
+    {
+	SN_CAUGHT(e);
+
+	// If snapper was just updated and the old snapperd is still
+	// running it might not know the GetDefaultSnapshot and
+	// GetActiveSnapshot methods.
+
+	if (strcmp(e.name(), "error.unknown_method") != 0)
+	    SN_RETHROW(e);
+    }
+
     if (list_mode != LM_ALL && list_mode != LM_SINGLE)
 	show_used_space = false;
 
@@ -460,13 +482,19 @@ list_from_one_config(ProxySnapper* snapper, ListMode list_mode, bool show_used_s
 	}
     }
 
-    auto add_date = [](TableRow& row, const ProxySnapshot& snapshot) {
-	row.add(snapshot.isCurrent() ? "" : datetime(snapshot.getDate(), utc, iso));
+    auto format_num = [&snapshots, default_snapshot, active_snapshot](const ProxySnapshot& snapshot) -> string {
+	bool is_default = default_snapshot != snapshots.end() && default_snapshot->getNum() == snapshot.getNum();
+	bool is_active = active_snapshot != snapshots.end() && active_snapshot->getNum() == snapshot.getNum();
+	static const char sign[2][2] = { { ' ', '-' }, { '+', '*' } };
+	return decString(snapshot.getNum()) + sign[is_default][is_active];
     };
 
-    auto add_used_space = [show_used_space](TableRow& row, const ProxySnapshot& snapshot) {
-	if (show_used_space)
-	    row.add(snapshot.isCurrent() ? "" : byte_to_humanstring(snapshot.getUsedSpace(), 2));
+    auto format_date = [](const ProxySnapshot& snapshot) -> string {
+	return snapshot.isCurrent() ? "" : datetime(snapshot.getDate(), utc, iso);
+    };
+
+    auto format_used_space = [](const ProxySnapshot& snapshot) -> string {
+	return snapshot.isCurrent() ? "" : byte_to_humanstring(snapshot.getUsedSpace(), 2);
     };
 
     Table table;
@@ -476,28 +504,28 @@ list_from_one_config(ProxySnapper* snapper, ListMode list_mode, bool show_used_s
 	case LM_ALL:
 	{
 	    TableHeader header;
+	    header.add(_("#"), TableAlign::RIGHT);
 	    header.add(_("Type"));
-	    header.add(_("#"));
-	    header.add(_("Pre #"));
+	    header.add(_("Pre #"), TableAlign::RIGHT);
 	    header.add(_("Date"));
 	    header.add(_("User"));
 	    if (show_used_space)
-		header.add(_("Used Space"));
+		header.add(_("Used Space"), TableAlign::RIGHT);
 	    header.add(_("Cleanup"));
 	    header.add(_("Description"));
 	    header.add(_("Userdata"));
 	    table.setHeader(header);
 
-	    const ProxySnapshots& snapshots = snapper->getSnapshots();
 	    for (const ProxySnapshot& snapshot : snapshots)
 	    {
 		TableRow row;
+		row.add(format_num(snapshot));
 		row.add(toString(snapshot.getType()));
-		row.add(decString(snapshot.getNum()));
 		row.add(snapshot.getType() == POST ? decString(snapshot.getPreNum()) : "");
-		add_date(row, snapshot);
+		row.add(format_date(snapshot));
 		row.add(username(snapshot.getUid()));
-		add_used_space(row, snapshot);
+		if (show_used_space)
+		    row.add(format_used_space(snapshot));
 		row.add(snapshot.getCleanup());
 		row.add(snapshot.getDescription());
 		row.add(show_userdata(snapshot.getUserdata()));
@@ -509,26 +537,26 @@ list_from_one_config(ProxySnapper* snapper, ListMode list_mode, bool show_used_s
 	case LM_SINGLE:
 	{
 	    TableHeader header;
-	    header.add(_("#"));
+	    header.add(_("#"), TableAlign::RIGHT);
 	    header.add(_("Date"));
 	    header.add(_("User"));
 	    if (show_used_space)
-		header.add(_("Used Space"));
+		header.add(_("Used Space"), TableAlign::RIGHT);
 	    header.add(_("Description"));
 	    header.add(_("Userdata"));
 	    table.setHeader(header);
 
-	    const ProxySnapshots& snapshots = snapper->getSnapshots();
 	    for (const ProxySnapshot& snapshot : snapshots)
 	    {
 		if (snapshot.getType() != SINGLE)
 		    continue;
 
 		TableRow row;
-		row.add(decString(snapshot.getNum()));
-		add_date(row, snapshot);
+		row.add(format_num(snapshot));
+		row.add(format_date(snapshot));
 		row.add(username(snapshot.getUid()));
-		add_used_space(row, snapshot);
+		if (show_used_space)
+		    row.add(format_used_space(snapshot));
 		row.add(snapshot.getDescription());
 		row.add(show_userdata(snapshot.getUserdata()));
 		table.add(row);
@@ -539,15 +567,14 @@ list_from_one_config(ProxySnapper* snapper, ListMode list_mode, bool show_used_s
 	case LM_PRE_POST:
 	{
 	    TableHeader header;
-	    header.add(_("Pre #"));
-	    header.add(_("Post #"));
+	    header.add(_("Pre #"), TableAlign::RIGHT);
+	    header.add(_("Post #"), TableAlign::RIGHT);
 	    header.add(_("Pre Date"));
 	    header.add(_("Post Date"));
 	    header.add(_("Description"));
 	    header.add(_("Userdata"));
 	    table.setHeader(header);
 
-	    const ProxySnapshots& snapshots = snapper->getSnapshots();
 	    for (ProxySnapshots::const_iterator it1 = snapshots.begin(); it1 != snapshots.end(); ++it1)
 	    {
 		if (it1->getType() != PRE)
@@ -561,8 +588,8 @@ list_from_one_config(ProxySnapper* snapper, ListMode list_mode, bool show_used_s
 		const ProxySnapshot& post = *it2;
 
 		TableRow row;
-		row.add(decString(pre.getNum()));
-		row.add(decString(post.getNum()));
+		row.add(format_num(pre));
+		row.add(format_num(post));
 		row.add(datetime(pre.getDate(), utc, iso));
 		row.add(datetime(post.getDate(), utc, iso));
 		row.add(pre.getDescription());
@@ -591,6 +618,9 @@ help_create()
 	 << _("\t--cleanup-algorithm, -c <algo>\tCleanup algorithm for snapshot.") << '\n'
 	 << _("\t--userdata, -u <userdata>\tUserdata for snapshot.") << '\n'
 	 << _("\t--command <command>\t\tRun command and create pre and post snapshots.") << endl
+	 << _("\t--read-only\t\t\tCreate read-only snapshot.") << '\n'
+	 << _("\t--read-write\t\t\tCreate read-write snapshot.") << '\n'
+	 << _("\t--from\t\t\t\tCreate a snapshot from the specified snapshot.") << '\n'
 	 << endl;
 }
 
@@ -606,6 +636,9 @@ command_create(ProxySnappers* snappers, ProxySnapper* snapper)
 	{ "cleanup-algorithm",	required_argument,	0,	'c' },
 	{ "userdata",		required_argument,	0,	'u' },
 	{ "command",		required_argument,	0,	0 },
+	{ "read-only",		no_argument,		0,	0 },
+	{ "read-write",		no_argument,		0,	0 },
+	{ "from",		required_argument,	0,	0 },
 	{ 0, 0, 0, 0 }
     };
 
@@ -626,6 +659,7 @@ command_create(ProxySnappers* snappers, ProxySnapper* snapper)
     bool print_number = false;
     SCD scd;
     string command;
+    ProxySnapshots::const_iterator parent = snapshots.getCurrent();
 
     GetOpts::parsed_opts::const_iterator opt;
 
@@ -667,6 +701,15 @@ command_create(ProxySnappers* snappers, ProxySnapper* snapper)
 	type = CT_PRE_POST;
     }
 
+    if ((opt = opts.find("read-only")) != opts.end())
+	scd.read_only = true;
+
+    if ((opt = opts.find("read-write")) != opts.end())
+	scd.read_only = false;
+
+    if ((opt = opts.find("from")) != opts.end())
+	parent = snapshots.findNum(opt->second);
+
     if (type == CT_POST && snapshot1 == snapshots.end())
     {
 	cerr << _("Missing or invalid pre-number.") << endl;
@@ -679,10 +722,22 @@ command_create(ProxySnappers* snappers, ProxySnapper* snapper)
 	exit(EXIT_FAILURE);
     }
 
+    if (type != CT_SINGLE && !scd.read_only)
+    {
+	cerr << _("Option --read-write only supported for snapshots of type single.") << endl;
+	exit(EXIT_FAILURE);
+    }
+
+    if (type != CT_SINGLE && parent != snapshots.getCurrent())
+    {
+	cerr << _("Option --from only supported for snapshots of type single.") << endl;
+	exit(EXIT_FAILURE);
+    }
+
     switch (type)
     {
 	case CT_SINGLE: {
-	    snapshot1 = snapper->createSingleSnapshot(scd);
+	    snapshot1 = snapper->createSingleSnapshot(parent, scd);
 	    if (print_number)
 		cout << snapshot1->getNum() << endl;
 	} break;
@@ -777,6 +832,37 @@ help_delete()
 }
 
 
+void
+filter_undeletables(ProxySnapshots& snapshots, vector<ProxySnapshots::iterator>& nums)
+{
+    auto filter = [&snapshots, &nums](ProxySnapshots::const_iterator undeletable, const char* message)
+    {
+	if (undeletable == snapshots.end())
+	    return;
+
+	unsigned int num = undeletable->getNum();
+
+	vector<ProxySnapshots::iterator>::iterator keep = find_if(nums.begin(), nums.end(),
+	    [num](ProxySnapshots::iterator it){ return num == it->getNum(); });
+
+	if (keep != nums.end())
+	{
+	    cerr << sformat(message, num) << endl;
+	    nums.erase(keep);
+	}
+    };
+
+    ProxySnapshots::const_iterator current_snapshot = snapshots.begin();
+    filter(current_snapshot, _("Cannot delete snapshot %d since it is the current system."));
+
+    ProxySnapshots::const_iterator active_snapshot = snapshots.getActive();
+    filter(active_snapshot, _("Cannot delete snapshot %d since it is the currently mounted snapshot."));
+
+    ProxySnapshots::const_iterator default_snapshot = snapshots.getDefault();
+    filter(default_snapshot, _("Cannot delete snapshot %d since it is the next to be mounted snapshot."));
+}
+
+
 void
 command_delete(ProxySnappers* snappers, ProxySnapper* snapper)
 {
@@ -833,6 +919,8 @@ command_delete(ProxySnappers* snappers, ProxySnapper* snapper)
 	}
     }
 
+    filter_undeletables(snapshots, nums);
+
     snapper->deleteSnapshots(nums, verbose);
 
     if (sync)
@@ -1215,10 +1303,12 @@ command_rollback(ProxySnappers* snappers, ProxySnapper* snapper)
 	exit(EXIT_FAILURE);
     }
 
+    const string default_description = "rollback backup";
+
     bool print_number = false;
 
     SCD scd1;
-    scd1.description = "rollback backup";
+    scd1.description = default_description;
     scd1.cleanup = "number";
     scd1.userdata["important"] = "yes";
 
@@ -1258,13 +1348,13 @@ command_rollback(ProxySnappers* snappers, ProxySnapper* snapper)
 	exit(EXIT_FAILURE);
     }
 
-    pair<bool, unsigned int> previous_default = filesystem->getDefault();
-
-    if (previous_default.first && scd1.description == "rollback backup")
-	scd1.description += sformat(" of #%d", previous_default.second);
-
     ProxySnapshots& snapshots = snapper->getSnapshots();
 
+    ProxySnapshots::iterator previous_default = snapshots.getDefault();
+
+    if (previous_default != snapshots.end() && scd1.description == default_description)
+        scd1.description += sformat(" of #%d", previous_default->getNum());
+
     ProxySnapshots::const_iterator snapshot1 = snapshots.end();
     ProxySnapshots::const_iterator snapshot2 = snapshots.end();
 
@@ -1310,15 +1400,11 @@ command_rollback(ProxySnappers* snappers, ProxySnapper* snapper)
 	    cout << " " << sformat(_("(Snapshot %d.)"), snapshot2->getNum()) << endl;
     }
 
-    if (previous_default.first)
+    if (previous_default != snapshots.end() && previous_default->getCleanup().empty())
     {
-	ProxySnapshots::iterator it = snapshots.find(previous_default.second);
-	if (it->getCleanup().empty())
-	{
-	    SMD smd = it->getSmd();
-	    smd.cleanup = "number";
-	    snapper->modifySnapshot(it, smd);
-	}
+	SMD smd = previous_default->getSmd();
+	smd.cleanup = "number";
+	snapper->modifySnapshot(previous_default, smd);
     }
 
     if (!quiet)
@@ -1790,6 +1876,12 @@ main(int argc, char** argv)
 	cerr << sformat(_("Quota error (%s)."), e.what()) << endl;
 	exit(EXIT_FAILURE);
     }
+    catch (const FreeSpaceException& e)
+    {
+	SN_CAUGHT(e);
+	cerr << sformat(_("Free space error (%s)."), e.what()) << endl;
+	exit(EXIT_FAILURE);
+    }
     catch (const Exception& e)
     {
 	SN_CAUGHT(e);
